from __future__ import absolute_import
from __future__ import unicode_literals

import time

from . import mock_socket

from nsq._compat import struct_h, struct_l, struct_q
from nsq import protocol, sync
sync.socket = mock_socket


def mock_write(c, data):
    c.s.queue_recv(data)


def mock_response_write(c, frame_type, data):
    body_size = 4 + len(data)
    body_size_packed = struct_l.pack(body_size)
    frame_type_packed = struct_l.pack(frame_type)
    mock_write(c, body_size_packed + frame_type_packed + data)


def mock_response_write_message(c, timestamp, attempts, id, body):
    timestamp_packed = struct_q.pack(timestamp)
    attempts_packed = struct_h.pack(attempts)
    id = ("%016d" % id).encode()
    mock_response_write(
        c, protocol.FRAME_TYPE_MESSAGE, timestamp_packed + attempts_packed + id + body)


def test_sync_authenticate_subscribe():
    c = sync.SyncConn()
    c.connect("127.0.0.1", 4150)

    c.send(protocol.identify({'short_id': 'test', 'long_id': 'test.example',
                              'client_id': 'test', 'hostname': 'test.example'}))
    c.send(protocol.subscribe('test', 'ch'))

    mock_response_write(c, protocol.FRAME_TYPE_RESPONSE, b'OK')
    mock_response_write(c, protocol.FRAME_TYPE_RESPONSE, b'OK')

    resp = c.read_response()
    unpacked = protocol.unpack_response(resp)
    assert unpacked[0] == protocol.FRAME_TYPE_RESPONSE
    assert unpacked[1] == b'OK'

    resp = c.read_response()
    unpacked = protocol.unpack_response(resp)
    assert unpacked[0] == protocol.FRAME_TYPE_RESPONSE
    assert unpacked[1] == b'OK'


def test_sync_receive_messages():
    c = sync.SyncConn()
    c.connect("127.0.0.1", 4150)

    c.send(protocol.identify({'short_id': 'test', 'long_id': 'test.example',
                              'client_id': 'test', 'hostname': 'test.example'}))
    c.send(protocol.subscribe('test', 'ch'))

    mock_response_write(c, protocol.FRAME_TYPE_RESPONSE, b'OK')
    mock_response_write(c, protocol.FRAME_TYPE_RESPONSE, b'OK')

    resp = c.read_response()
    unpacked = protocol.unpack_response(resp)
    assert unpacked[0] == protocol.FRAME_TYPE_RESPONSE
    assert unpacked[1] == b'OK'

    resp = c.read_response()
    unpacked = protocol.unpack_response(resp)
    assert unpacked[0] == protocol.FRAME_TYPE_RESPONSE
    assert unpacked[1] == b'OK'

    for i in range(10):
        c.send(protocol.ready(1))
        body = ('{"data": {"test_key": %d}}' % i).encode()
        ts = int(time.time() * 1000 * 1000)
        mock_response_write_message(c, ts, 0, i, body)
        resp = c.read_response()
        unpacked = protocol.unpack_response(resp)
        assert unpacked[0] == protocol.FRAME_TYPE_MESSAGE
        msg = protocol.decode_message(unpacked[1])
        assert msg.timestamp == ts
        assert msg.id == ("%016d" % i).encode()
        assert msg.attempts == 0
        assert msg.body == body



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import signal
import subprocess
import time
import ssl

import tornado.gen
import tornado.httpclient
import tornado.httpserver
import tornado.testing
import tornado.web

# shunt '..' into sys.path since we are in a 'tests' subdirectory
base_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
if base_dir not in sys.path:
    sys.path.insert(0, base_dir)

from nsq import protocol
from nsq.conn import AsyncConn
from nsq.reader import Reader
from nsq.deflate_socket import DeflateSocket
from nsq.snappy_socket import SnappySocket


class IntegrationBase(tornado.testing.AsyncTestCase):
    nsqd_command = []
    nsqlookupd_command = []

    def setUp(self):
        super(IntegrationBase, self).setUp()
        if not hasattr(self, 'processes'):
            self.processes = []

        if self.nsqlookupd_command:
            proc = subprocess.Popen(self.nsqlookupd_command)
            self.processes.append(proc)
            self.wait_ping('http://127.0.0.1:4161/ping')
        if self.nsqd_command:
            proc = subprocess.Popen(self.nsqd_command)
            self.processes.append(proc)
            self.wait_ping('http://127.0.0.1:4151/ping')

    def tearDown(self):
        super(IntegrationBase, self).tearDown()
        for proc in self.processes:
            os.kill(proc.pid, signal.SIGKILL)
            proc.wait()

    def wait_ping(self, endpoint):
        start = time.time()
        http = tornado.httpclient.HTTPClient()
        while True:
            try:
                resp = http.fetch(endpoint)
                if resp.body == b'OK':
                    break
                continue
            except Exception:
                if time.time() - start > 5:
                    raise
                time.sleep(0.1)
                continue

    def _send_messages(self, topic, count, body):
        c = AsyncConn('127.0.0.1', 4150)
        c.connect()

        def _on_ready(*args, **kwargs):
            for i in range(count):
                c.send(protocol.pub(topic, body))

        c.on('ready', _on_ready)


class ReaderIntegrationTest(IntegrationBase):
    identify_options = {
        'user_agent': 'sup',
        'snappy': True,
        'tls_v1': True,
        'tls_options': {'cert_reqs': ssl.CERT_NONE},
        'heartbeat_interval': 10,
        'output_buffer_size': 4096,
        'output_buffer_timeout': 50
    }

    nsqd_command = ['nsqd', '--verbose', '--snappy',
                    '--tls-key=%s/tests/key.pem' % base_dir,
                    '--tls-cert=%s/tests/cert.pem' % base_dir]

    def test_bad_reader_arguments(self):
        topic = 'test_reader_msgs_%s' % time.time()
        bad_options = dict(self.identify_options)
        bad_options.update(dict(foo=10))

        def handler(msg):
            return None

        self.assertRaises(
            AssertionError,
            Reader,
            nsqd_tcp_addresses=['127.0.0.1:4150'],
            topic=topic, channel='ch',
            message_handler=handler, max_in_flight=100,
            **bad_options)

    def test_conn_identify(self):
        c = AsyncConn('127.0.0.1', 4150)
        c.on('identify_response', self.stop)
        c.connect()
        response = self.wait()
        print(response)
        assert response['conn'] is c
        assert isinstance(response['data'], dict)

    def test_conn_identify_options(self):
        c = AsyncConn('127.0.0.1', 4150, **self.identify_options)
        c.on('identify_response', self.stop)
        c.connect()
        response = self.wait()
        print(response)
        assert response['conn'] is c
        assert isinstance(response['data'], dict)
        assert response['data']['snappy'] is True
        assert response['data']['tls_v1'] is True

    def test_conn_socket_upgrade(self):
        c = AsyncConn('127.0.0.1', 4150, **self.identify_options)
        c.on('ready', self.stop)
        c.connect()
        self.wait()
        assert isinstance(c.socket, SnappySocket)
        assert isinstance(c.socket._socket, ssl.SSLSocket)

    def test_conn_subscribe(self):
        topic = 'test_conn_suscribe_%s' % time.time()
        c = AsyncConn('127.0.0.1', 4150, **self.identify_options)

        def _on_ready(*args, **kwargs):
            c.on('response', self.stop)
            c.send(protocol.subscribe(topic, 'ch'))

        c.on('ready', _on_ready)
        c.connect()
        response = self.wait()
        print(response)
        assert response['conn'] is c
        assert response['data'] == b'OK'

    def test_conn_messages(self):
        self.msg_count = 0

        topic = 'test_conn_suscribe_%s' % time.time()
        self._send_messages(topic, 5, b'sup')

        c = AsyncConn('127.0.0.1', 4150, **self.identify_options)

        def _on_message(*args, **kwargs):
            self.msg_count += 1
            if c.in_flight == 5:
                self.stop()

        def _on_ready(*args, **kwargs):
            c.on('message', _on_message)
            c.send(protocol.subscribe(topic, 'ch'))
            c.send_rdy(5)

        c.on('ready', _on_ready)
        c.connect()

        self.wait()
        assert self.msg_count == 5

    def test_reader_messages(self):
        self.msg_count = 0
        num_messages = 500

        topic = 'test_reader_msgs_%s' % time.time()
        self._send_messages(topic, num_messages, b'sup')

        def handler(msg):
            assert msg.body == b'sup'
            self.msg_count += 1
            if self.msg_count >= num_messages:
                self.stop()
            return True

        r = Reader(nsqd_tcp_addresses=['127.0.0.1:4150'], topic=topic, channel='ch',
                   message_handler=handler, max_in_flight=100,
                   **self.identify_options)

        self.wait()
        r.close()
        assert self.msg_count == num_messages

    def test_reader_coro(self):
        self.msg_count = 0
        num_messages = 20

        topic = 'test_reader_msgs_%s' % time.time()
        self._send_messages(topic, num_messages, b'sup')

        @tornado.gen.coroutine
        def handler(msg):
            yield tornado.gen.sleep(0.1)
            self.msg_count += 1
            if self.msg_count >= num_messages:
                self.stop()
            raise tornado.gen.Return(True)

        r = Reader(nsqd_tcp_addresses=['127.0.0.1:4150'], topic=topic, channel='ch',
                   message_handler=handler, max_in_flight=10,
                   **self.identify_options)

        self.wait()
        r.close()
        assert self.msg_count == num_messages

    def test_reader_heartbeat(self):
        this = self
        this.count = 0

        def handler(msg):
            return True

        class HeartbeatReader(Reader):
            def heartbeat(self, conn):
                this.count += 1
                if this.count == 2:
                    this.stop()

        topic = 'test_reader_hb_%s' % time.time()
        HeartbeatReader(nsqd_tcp_addresses=['127.0.0.1:4150'], topic=topic, channel='ch',
                        message_handler=handler, max_in_flight=100,
                        heartbeat_interval=1)
        self.wait()


class DeflateReaderIntegrationTest(IntegrationBase):
    identify_options = {
        'user_agent': 'sup',
        'deflate': True,
        'deflate_level': 6,
        'tls_options': {'cert_reqs': ssl.CERT_NONE},
        'heartbeat_interval': 10,
        'output_buffer_size': 4096,
        'output_buffer_timeout': 50
    }

    nsqd_command = ['nsqd', '--verbose', '--deflate',
                    '--tls-key=%s/tests/key.pem' % base_dir,
                    '--tls-cert=%s/tests/cert.pem' % base_dir]

    def test_conn_identify_options(self):
        c = AsyncConn('127.0.0.1', 4150, **self.identify_options)
        c.on('identify_response', self.stop)
        c.connect()
        response = self.wait()
        print(response)
        assert response['conn'] is c
        assert isinstance(response['data'], dict)
        assert response['data']['deflate'] is True

    def test_conn_socket_upgrade(self):
        c = AsyncConn('127.0.0.1', 4150, **self.identify_options)
        c.on('ready', self.stop)
        c.connect()
        self.wait()
        assert isinstance(c.socket, DeflateSocket)

    def test_reader_messages(self):
        self.msg_count = 0
        num_messages = 300
        topic = 'test_reader_msgs_%s' % time.time()
        self._send_messages(topic, num_messages, b'sup')

        def handler(msg):
            assert msg.body == b'sup'
            self.msg_count += 1
            if self.msg_count >= num_messages:
                self.stop()
            return True

        r = Reader(nsqd_tcp_addresses=['127.0.0.1:4150'],
                   topic=topic, channel='ch',
                   message_handler=handler, max_in_flight=100,
                   **self.identify_options)
        self.wait()
        r.close()


class LookupdReaderIntegrationTest(IntegrationBase):
    nsqd_command = ['nsqd', '--verbose', '-lookupd-tcp-address=127.0.0.1:4160']
    nsqlookupd_command = ['nsqlookupd', '--verbose', '-broadcast-address=127.0.0.1']

    identify_options = {
        'user_agent': 'sup',
        'heartbeat_interval': 10,
        'output_buffer_size': 4096,
        'output_buffer_timeout': 50
    }

    def test_reader_messages(self):
        self.msg_count = 0
        num_messages = 1

        topic = 'test_lookupd_msgs_%s' % int(time.time())
        self._send_messages(topic, num_messages, b'sup')

        def handler(msg):
            assert msg.body == b'sup'
            self.msg_count += 1
            if self.msg_count >= num_messages:
                self.stop()
            return True

        r = Reader(lookupd_http_addresses=['http://127.0.0.1:4161'],
                   topic=topic,
                   channel='ch',
                   message_handler=handler,
                   lookupd_poll_interval=1,
                   lookupd_poll_jitter=0.01,
                   max_in_flight=100,
                   **self.identify_options)

        self.wait()
        r.close()


class AuthHandler(tornado.web.RequestHandler):
    def get(self):
        if self.get_argument('secret') == 'opensesame':
            self.finish({
                'ttl': 30,
                'identity': 'username',
                'authorizations': [{
                    'permissions': ['subscribe'],
                    'topic': 'authtopic',
                    'channels': ['ch'],
                }]
            })
        else:
            self.set_status(403)


class ReaderAuthIntegrationTest(IntegrationBase):
    identify_options = {
        'user_agent': 'supauth',
        'heartbeat_interval': 10,
        'output_buffer_size': 4096,
        'output_buffer_timeout': 50,
        'auth_secret': "opensesame",
    }

    def setUp(self):
        auth_sock, auth_port = tornado.testing.bind_unused_port()
        self.auth_sock = auth_sock
        self.nsqd_command = [
            'nsqd', '--verbose', '--auth-http-address=127.0.0.1:%d' % auth_port
        ]
        super(ReaderAuthIntegrationTest, self).setUp()

    def tearDown(self):
        super(ReaderAuthIntegrationTest, self).tearDown()
        self.auth_sock.close()

    def test_conn_identify(self):
        auth_app = tornado.web.Application([("/auth", AuthHandler)])
        auth_srv = tornado.httpserver.HTTPServer(auth_app)
        auth_srv.add_socket(self.auth_sock)

        c = AsyncConn('127.0.0.1', 4150, **self.identify_options)

        def _on_ready(*args, **kwargs):
            c.on('response', self.stop)
            c.send(protocol.subscribe('authtopic', 'ch'))

        c.on('ready', _on_ready)
        c.connect()
        response = self.wait()
        auth_srv.stop()
        print(response)
        assert response['conn'] is c
        assert response['data'] == b'OK'



from __future__ import absolute_import

import collections
import time

from mock import patch, create_autospec
from tornado.ioloop import IOLoop
from tornado.concurrent import Future

import nsq
from nsq import event

_conn_port = 4150


def get_reader(max_in_flight=5):
    r = nsq.Reader(
        "test", "test",
        message_handler=_message_handler,
        lookupd_http_addresses=["http://localhost:4161"],
        max_in_flight=max_in_flight,
        max_backoff_duration=2.0,
    )
    r.conns = collections.OrderedDict()
    return r


def get_ioloop():
    ioloop = create_autospec(IOLoop, instance=True)
    ioloop.time.side_effect = time.time
    ioloop.call_later.side_effect = lambda dt, cb: ioloop.add_timeout(time.time() + dt, cb)
    ioloop.add_timeout.side_effect = lambda _, cb: cb
    return ioloop


def get_conn(reader):
    global _conn_port
    with patch('nsq.conn.IOStream', autospec=True) as iostream:
        instance = iostream.return_value
        instance.connect.return_value = Future()
        instance.read_bytes.return_value = Future()
        conn = reader.connect_to_nsqd('localhost', _conn_port)
    _conn_port += 1
    conn.trigger(event.READY, conn=conn)
    return conn


def send_message(conn):
    msg = _get_message(conn)
    conn.in_flight += 1
    conn.trigger(event.MESSAGE, conn=conn, message=msg)
    return msg


def _get_message(conn):
    msg = nsq.Message("1234", "{}", 1234, 0)
    msg.on('finish', conn._on_message_finish)
    msg.on('requeue', conn._on_message_requeue)
    return msg


def _message_handler(msg):
    msg.enable_async()



import pytest
from mock import patch

from nsq import event

from .reader_unit_test_helpers import (
    get_reader,
    get_conn,
    get_ioloop,
    send_message
)


def test_initial_probe_then_full_throttle():
    r = get_reader()
    conn = get_conn(r)
    assert conn.rdy == 1
    send_message(conn)
    assert conn.rdy == r._connection_max_in_flight()


def test_new_conn_throttles_down_existing_conns():
    r = get_reader()
    conn1 = get_conn(r)
    send_message(conn1)
    assert conn1.rdy == r._connection_max_in_flight()

    conn2 = get_conn(r)
    assert conn2.rdy == 1
    assert conn1.rdy == r._connection_max_in_flight()


def test_new_conn_respects_max_in_flight():
    max_in_flight = 1
    r = get_reader(max_in_flight)
    get_conn(r)
    get_conn(r)


@pytest.mark.parametrize(['conn_count', 'max_in_flight'], [
    [5, 10],
    [5, 4],
])
@patch("tornado.ioloop.IOLoop.current")
def test_rdy_complete_backoff(ioloop_current_mock, conn_count, max_in_flight):
    ioloop_mock = get_ioloop()
    ioloop_current_mock.return_value = ioloop_mock
    r = get_reader(max_in_flight=max_in_flight)
    conns = [get_conn(r) for _ in range(conn_count)]

    msg = send_message(conns[0])
    msg.trigger(event.REQUEUE, message=msg)

    timeout_args, _ = ioloop_mock.add_timeout.call_args
    conn = timeout_args[1]()
    assert r.total_rdy == 1

    msg = send_message(conn)
    msg.trigger(event.FINISH, message=msg)
    assert r.total_rdy == max_in_flight


@pytest.mark.parametrize(['conn_count', 'old_max', 'new_max'], [
    [4, 8, 12],
    [4, 3, 12],
    [4, 2, 3],
    [4, 12, 8],
    [4, 12, 3],
    [4, 3, 2],
])
def test_rdy_set_max_in_flight(old_max, new_max, conn_count):
    r = get_reader(old_max)
    conns = [get_conn(r) for _ in range(conn_count)]
    for c in conns:
        msg = send_message(c)
        msg.trigger(event.FINISH, message=msg)
    r.set_max_in_flight(new_max)
    assert r.total_rdy <= new_max
    for c in conns:
        msg = send_message(c)
        msg.trigger(event.FINISH, message=msg)
    assert r.total_rdy == new_max


def test_rdy_disable_enable():
    max_in_flight = 10
    r = get_reader(max_in_flight)
    conns = [get_conn(r) for _ in range(max_in_flight // 2 - 1)]
    msg = send_message(conns[0])

    r.set_max_in_flight(0)
    assert r.total_rdy == 0

    conns.append(get_conn(r))
    assert r.total_rdy == 0

    msg.trigger(event.FINISH, message=msg)
    assert r.total_rdy == 0

    r.set_max_in_flight(max_in_flight)
    assert r.total_rdy == len(conns)


@patch("tornado.ioloop.IOLoop.current")
def test_rdy_disable_enable_in_backoff(ioloop_current_mock):
    ioloop_mock = get_ioloop()
    ioloop_current_mock.return_value = ioloop_mock
    max_in_flight = 10
    r = get_reader(max_in_flight)
    conns = [get_conn(r) for _ in range(2)]
    msg = send_message(conns[0])

    r.set_max_in_flight(0)
    assert r.total_rdy == 0

    msg.trigger(event.REQUEUE, message=msg)
    timeout_args, _ = ioloop_mock.add_timeout.call_args
    timeout_args[1]()
    assert r.total_rdy == 0

    r.set_max_in_flight(max_in_flight)
    assert r.total_rdy == 1


@patch("random.randrange")
@patch("tornado.ioloop.IOLoop.current")
def test_rdy_retry(ioloop_current_mock, randrange_mock):
    ioloop_mock = get_ioloop()
    ioloop_current_mock.return_value = ioloop_mock
    randrange_mock.return_value = 0
    r = get_reader(max_in_flight=1)
    get_conn(r)
    r.set_max_in_flight(0)
    c = get_conn(r)
    r.set_max_in_flight(1)
    if c.rdy_timeout is not None:
        timeout_args, _ = ioloop_mock.add_timeout.call_args
        timeout_args[1]()
    assert r.total_rdy <= r.max_in_flight


@patch("random.randrange")
@patch("time.time")
@patch("tornado.ioloop.IOLoop.current")
def test_rdy_redistribution_idle_conns_lose_rdy(ioloop_current_mock, time_mock, randrange_mock):
    ioloop_mock = get_ioloop()
    ioloop_current_mock.return_value = ioloop_mock
    t = 1
    time_mock.side_effect = lambda: t
    randrange_mock.return_value = 1
    r = get_reader(max_in_flight=2)
    conn1 = get_conn(r)
    conn2 = get_conn(r)
    conn3 = get_conn(r)
    assert conn1.rdy == conn2.rdy == 1
    assert conn3.rdy == 0
    msg = send_message(conn1)
    msg.trigger(event.FINISH, message=msg)
    t += r.low_rdy_idle_timeout - 0.1
    r._redistribute_rdy_state()
    assert conn1.rdy == conn2.rdy == 1
    assert conn3.rdy == 0
    t += 0.2
    r._redistribute_rdy_state()
    assert conn1.rdy == 0
    assert conn2.rdy == conn3.rdy == 1


@patch("time.time")
@patch("tornado.ioloop.IOLoop.current")
def test_rdy_redistribution_doesnt_reallocate_rdy_in_use(ioloop_current_mock, time_mock):
    ioloop_mock = get_ioloop()
    ioloop_current_mock.return_value = ioloop_mock
    t = 1
    time_mock.side_effect = lambda: t
    r = get_reader(max_in_flight=2)
    conn1 = get_conn(r)
    conn2 = get_conn(r)
    get_conn(r)
    assert r.total_rdy == 2
    msg1 = send_message(conn1)
    msg2 = send_message(conn2)
    msg1.trigger(event.FINISH, message=msg1)
    t += r.low_rdy_idle_timeout + 0.1
    r._redistribute_rdy_state()
    assert r.total_rdy == 1
    msg2.trigger(event.FINISH, message=msg2)
    r._redistribute_rdy_state()
    assert r.total_rdy == 2


@patch("random.randrange")
@patch("time.time")
@patch("tornado.ioloop.IOLoop.current")
def test_rdy_redistribution_isnt_pinned_to_conns(ioloop_current_mock, time_mock, randrange_mock):
    ioloop_mock = get_ioloop()
    ioloop_current_mock.return_value = ioloop_mock
    t = 1
    time_mock.side_effect = lambda: t
    randrange_mock.return_value = 1
    r = get_reader(max_in_flight=2)
    conn1 = get_conn(r)
    conn2 = get_conn(r)
    conn3 = get_conn(r)
    assert conn1.rdy == conn2.rdy == 1
    assert conn3.rdy == 0
    msg1 = send_message(conn1)
    msg2 = send_message(conn2)
    msg1.trigger(event.FINISH, message=msg1)
    t += 1
    r._redistribute_rdy_state()
    assert conn1.rdy == 1
    assert conn2.rdy == conn3.rdy == 0
    t += 1
    msg2.trigger(event.FINISH, message=msg2)
    r._redistribute_rdy_state()
    assert conn1.rdy == conn3.rdy == 1
    assert conn2.rdy == 0


@patch("random.choice")
@patch("random.randrange")
@patch("time.time")
@patch("tornado.ioloop.IOLoop.current")
def test_rdy_redistribution_idle_conns_lose_rdy_in_backoff(
        ioloop_current_mock, time_mock, randrange_mock, choice_mock):
    ioloop_mock = get_ioloop()
    ioloop_current_mock.return_value = ioloop_mock
    t = 1
    time_mock.side_effect = lambda: t
    randrange_mock.return_value = 2
    r = get_reader(max_in_flight=2)
    conn1 = get_conn(r)
    conn2 = get_conn(r)
    conn3 = get_conn(r)
    choice_mock.return_value = conn1
    assert conn1.rdy == conn2.rdy == 1
    assert conn3.rdy == 0
    msg = send_message(conn1)
    msg.trigger(event.REQUEUE, message=msg)
    timeout_args, _ = ioloop_mock.add_timeout.call_args
    timeout_args[1]()
    assert conn1.rdy == r.total_rdy == 1
    t += r.low_rdy_idle_timeout + 0.1
    r._redistribute_rdy_state()
    assert conn3.rdy == r.total_rdy == 1


@patch("random.choice")
@patch("time.time")
@patch("tornado.ioloop.IOLoop.current")
def test_rdy_redistribution_doesnt_reallocate_rdy_in_use_in_backoff(
        ioloop_current_mock, time_mock, choice_mock):
    ioloop_mock = get_ioloop()
    ioloop_current_mock.return_value = ioloop_mock
    t = 1
    time_mock.side_effect = lambda: t
    r = get_reader(max_in_flight=2)
    conn1 = get_conn(r)
    conn2 = get_conn(r)
    get_conn(r)
    assert r.total_rdy == 2
    msg = send_message(conn1)
    send_message(conn2)
    msg.trigger(event.REQUEUE, message=msg)
    timeout_args, _ = ioloop_mock.add_timeout.call_args
    choice_mock.return_value = conn1
    timeout_args[1]()
    assert r.total_rdy == 1
    t += r.low_rdy_idle_timeout + 0.1
    r._redistribute_rdy_state()
    assert r.total_rdy == 0



"""
Mock socket module, copied (and edited) from:
http://svn.python.org/projects/python/branches/pep-0384/Lib/test/mock_socket.py
"""


class MockSocket:
    def __init__(self):
        self.output = []
        self.lines = []
        self.conn = None
        self.timeout = None

    def connect(self, addr):
        pass

    def queue_recv(self, line):
        self.lines.append(line)

    def recv(self, bufsize, flags=None):
        return self.lines.pop(0)

    def settimeout(self, timeout=None):
        if timeout is None:
            self.timeout = 60
        else:
            self.timeout = timeout

    def gettimeout(self):
        return self.timeout

    def send(self, data, flags=None):
        self.output.append(data)
        return len(data)

    def close(self):
        pass

    def setblocking(self, val):
        pass


def socket(family=None, type=None, proto=None):
    return MockSocket()


# Constants
AF_INET = None
SOCK_STREAM = None
SOL_SOCKET = None
SO_REUSEADDR = None



from __future__ import absolute_import
from __future__ import unicode_literals

import os
import sys
import json

import pytest

# shunt '..' into sys.path since we are in a 'tests' subdirectory
base_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
if base_dir not in sys.path:
    sys.path.insert(0, base_dir)

from nsq._compat import to_bytes
from nsq._compat import struct_l
from nsq import protocol


identify_dict_ascii = {'a': 1, 'b': 2}
identify_dict_unicode = {'c': u'w\xc3\xa5\xe2\x80\xa0'}
identify_body_ascii = to_bytes(json.dumps(identify_dict_ascii))
identify_body_unicode = to_bytes(json.dumps(identify_dict_unicode))
msgs = [b'asdf', b'ghjk', b'abcd']
mpub_body = struct_l.pack(len(msgs)) + b''.join(struct_l.pack(len(m)) + m for m in msgs)


@pytest.mark.parametrize(['cmd_method', 'kwargs', 'result'], [
    pytest.param(protocol.identify,
                 {'data': identify_dict_ascii},
                 b'IDENTIFY\n' + struct_l.pack(len(identify_body_ascii))
                 + to_bytes(identify_body_ascii),  # noqa: W503
                 id="identify-ascii"),
    pytest.param(protocol.identify,
                 {'data': identify_dict_unicode},
                 b'IDENTIFY\n' + struct_l.pack(len(identify_body_unicode))
                 + to_bytes(identify_body_unicode),  # noqa: W503
                 id="identify-unicode"),
    pytest.param(protocol.subscribe,
                 {'topic': 'test_topic', 'channel': 'test_channel'},
                 b'SUB test_topic test_channel\n',
                 id="subscribe-topic"),
    pytest.param(protocol.finish,
                 {'id': 'test'},
                 b'FIN test\n',
                 id="finish-ascii"),
    pytest.param(protocol.finish,
                 {'id': u'\u2020est \xfcn\xee\xe7\xf8\u2202\xe9'},
                 b'FIN \xe2\x80\xa0est \xc3\xbcn\xc3\xae\xc3\xa7\xc3\xb8\xe2\x88\x82\xc3\xa9\n',
                 id="finish-binary"),
    pytest.param(protocol.requeue,
                 {'id': 'test'},
                 b'REQ test 0\n',
                 id="requeue-0"),
    pytest.param(protocol.requeue,
                 {'id': 'test', 'time_ms': 60},
                 b'REQ test 60\n',
                 id="requeue-60"),
    pytest.param(protocol.touch,
                 {'id': 'test'},
                 b'TOUCH test\n',
                 id="touch"),
    pytest.param(protocol.ready,
                 {'count': 100},
                 b'RDY 100\n',
                 id="ready"),
    pytest.param(protocol.nop,
                 {},
                 b'NOP\n',
                 id="nop"),
    pytest.param(protocol.pub,
                 {'topic': 'test', 'data': msgs[0]},
                 b'PUB test\n' + struct_l.pack(len(msgs[0])) + to_bytes(msgs[0]),
                 id="pub"),
    pytest.param(protocol.mpub,
                 {'topic': 'test', 'data': msgs},
                 b'MPUB test\n' + struct_l.pack(len(mpub_body)) + to_bytes(mpub_body),
                 id="mpub"),
])
def test_command(cmd_method, kwargs, result):
    assert cmd_method(**kwargs) == result


def test_unicode_body():
    pytest.raises(AssertionError, protocol.pub, 'topic', u'unicode body')



from __future__ import absolute_import

from decimal import Decimal
import os
import sys

import pytest

# shunt '..' into sys.path since we are in a 'tests' subdirectory
base_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
if base_dir not in sys.path:
    sys.path.insert(0, base_dir)

from nsq import BackoffTimer
from nsq import protocol


@pytest.mark.parametrize(['name', 'good'], [
    ('valid_name', True),
    ('invalid name with space', False),
    ('invalid_name_due_to_length_this_is_really_really_really_really_long', False),
    ('test-with_period.', True),
    ('test#ephemeral', True),
    ('test:ephemeral', False),
])
def test_topic_names(name, good):
    assert protocol.valid_topic_name(name) == good


@pytest.mark.parametrize(['name', 'good'], [
    ('test', True),
    ('test-with_period.', True),
    ('test#ephemeral', True),
    ('invalid_name_due_to_length_this_is_really_really_really_really_long', False),
    ('invalid name with space', False),
])
def test_channel_names(name, good):
    assert protocol.valid_channel_name(name) == good


def test_backoff_timer():
    timer = BackoffTimer(.1, 120, long_length=1000)
    assert timer.get_interval() == .1
    timer.success()
    assert timer.get_interval() == .1
    timer.failure()
    interval = '%0.2f' % timer.get_interval()
    assert interval == '3.19'
    assert timer.min_interval == Decimal('.1')
    assert timer.short_interval == Decimal('2.9975')
    assert timer.long_interval == Decimal('0.089925')

    timer.failure()
    interval = '%0.2f' % timer.get_interval()
    assert interval == '6.27'
    timer.success()
    interval = '%0.2f' % timer.get_interval()
    assert interval == '3.19'
    for i in range(25):
        timer.failure()
    interval = '%0.2f' % timer.get_interval()
    assert interval == '32.41'



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="yeansq", # yet another nsq
    version="0.0.1",  # Latest version .
    author="slipper",
    author_email="r2fscg@gmail.com",
    description="A package for faster coding.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/private_repo/nsq",
    packages=setuptools.find_packages(),
    install_requires=[
        'tornado'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from __future__ import absolute_import
import os
import sys

from mock import patch, create_autospec
from tornado.iostream import IOStream
from tornado.concurrent import Future

# shunt '..' into sys.path since we are in a 'tests' subdirectory
base_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
if base_dir not in sys.path:
    sys.path.insert(0, base_dir)

from nsq._compat import struct_l
from nsq.conn import AsyncConn
from nsq import protocol


# The functions below are meant to be used as specs for mocks of callbacks. Yay dynamic typing
def f(*args, **kwargs):
    pass


def _get_test_conn():
    conn = AsyncConn('test', 4150)
    # now set the stream attribute, which is ordinarily set in conn.connect()
    conn.stream = create_autospec(IOStream)
    fut = Future()
    fut.set_result(conn.stream)
    conn.stream.connect.return_value = fut
    conn.stream.read_bytes.return_value = Future()
    return conn


# We'd like to autospec the socket mock here, but due to a bug in mock on
# Python 2.x, we can't currently do that. See:
#
#   https://github.com/testing-cabal/mock/issues/323
#
@patch('nsq.conn.socket')
@patch('nsq.conn.IOStream', autospec=True)
def test_connect(mock_iostream, mock_socket):
    instance = mock_iostream.return_value
    fut = Future()
    fut.set_result(instance)
    instance.connect.return_value = fut

    conn = _get_test_conn()
    conn.connect()
    conn.stream.set_close_callback.assert_called_once_with(conn._socket_close)
    conn.stream.connect.assert_called_once_with((conn.host, conn.port))
    assert conn.state == 'CONNECTING'

    # now ensure that we will bail if we were already in a connecting or connected state
    conn = _get_test_conn()
    conn.state = 'CONNECTING'
    conn.connect()
    assert not conn.stream.set_close_callback.called
    assert not conn.stream.connect.called

    conn = _get_test_conn()
    conn.state = 'CONNECTED'
    conn.connect()
    assert not conn.stream.set_close_callback.called
    assert not conn.stream.connect.called


def test_connect_callback():
    conn = _get_test_conn()
    on_connect = create_autospec(f)
    conn.on('connect', on_connect)
    # simulate having called `conn.connect` by setting state to `connecting`
    conn.state = 'CONNECTING'
    with patch.object(conn, '_start_read', autospec=True) as mock_start_read:
        fut = Future()
        fut.set_result(conn.stream)
        conn._connect_callback(fut)
        assert conn.state == 'CONNECTED'
        conn.stream.write.assert_called_once_with(protocol.MAGIC_V2)
        mock_start_read.assert_called_once_with()
        on_connect.assert_called_once_with(conn=conn)


def test_start_read():
    conn = _get_test_conn()
    conn._start_read()
    conn.stream.read_bytes.assert_called_once_with(4)


def test_read_size():
    conn = _get_test_conn()
    body_size = 6
    body_size_packed = struct_l.pack(body_size)
    fut = Future()
    fut.set_result(body_size_packed)
    conn._read_size(fut)
    conn.stream.read_bytes.assert_called_once_with(body_size)

    # now test that we get the right behavior when we get malformed data
    # for this, we'll want to stick on mock on conn.close
    conn.stream.read_bytes.reset_mock()
    with patch.object(conn, 'close', autospec=True) as mock_close:
        fut = Future()
        fut.set_result('asdfasdf')
        conn._read_size(fut)
        mock_close.assert_called_once_with()
        assert not conn.stream.read_bytes.called


def test_read_body():
    conn = _get_test_conn()
    on_data = create_autospec(f)
    conn.on('data', on_data)

    data = 'NSQ'
    fut = Future()
    fut.set_result(data)
    conn._read_body(fut)
    on_data.assert_called_once_with(conn=conn, data=data)
    conn.stream.read_bytes.assert_called_once_with(4)

    # now test functionality when the data_callback fails
    on_data.reset_mock()
    conn.stream.read_bytes.reset_mock()
    on_data.return_value = Exception("Boom.")
    conn._read_body(fut)
    # verify that the next _start_read was called
    conn.stream.read_bytes.assert_called_once_with(4)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import random
import time

from mock import call, patch
from tornado import gen
from tornado.testing import AsyncTestCase, gen_test

from .reader_unit_test_helpers import (
    get_reader,
    get_ioloop,
    get_conn,
    send_message
)

# shunt '..' into sys.path since we are in a 'tests' subdirectory
base_dir = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
if base_dir not in sys.path:
    sys.path.insert(0, base_dir)

from nsq import event


class BackoffTests(AsyncTestCase):

    @gen_test
    def test_backoff_easy(self):
        r = get_reader()
        conn = get_conn(r)

        msg = send_message(conn)

        msg.trigger(event.FINISH, message=msg)
        assert r.backoff_block is False
        assert r.backoff_timer.get_interval() == 0

        msg = send_message(conn)

        msg.trigger(event.REQUEUE, message=msg)
        assert r.backoff_block is True
        assert r.backoff_timer.get_interval() > 0

        yield gen.sleep(r.backoff_timer.get_interval() + 0.05)

        assert r.backoff_block is False
        send_args, send_kwargs = conn.stream.write.call_args
        assert send_args[0] == b'RDY 1\n'

        msg = send_message(conn)

        msg.trigger(event.FINISH, message=msg)
        assert r.backoff_block is False
        assert r.backoff_timer.get_interval() == 0

        expected_args = [
            b'SUB test test\n',
            b'RDY 1\n',
            b'RDY 5\n',
            b'FIN 1234\n',
            b'RDY 0\n',
            b'REQ 1234 0\n',
            b'RDY 1\n',
            b'RDY 5\n',
            b'FIN 1234\n'
        ]
        assert conn.stream.write.call_args_list == [call(arg) for arg in expected_args]

    @gen_test
    def test_backoff_out_of_order(self):
        r = get_reader(max_in_flight=4)
        conn1 = get_conn(r)
        conn2 = get_conn(r)

        msg = send_message(conn1)

        msg.trigger(event.FINISH, message=msg)
        assert r.backoff_block is False
        assert r.backoff_timer.get_interval() == 0

        msg = send_message(conn1)

        msg.trigger(event.REQUEUE, message=msg)
        assert r.backoff_block is True
        assert r.backoff_timer.get_interval() > 0
        backoff_interval = r.backoff_timer.get_interval()

        msg = send_message(conn1)

        msg.trigger(event.FINISH, message=msg)
        assert r.backoff_block is True
        assert r.backoff_timer.get_interval() == 0

        yield gen.sleep(backoff_interval + 0.05)

        assert r.backoff_block is False
        assert r.backoff_timer.get_interval() == 0

        expected_args = [
            b'SUB test test\n',
            b'RDY 1\n',
            b'RDY 2\n',
            b'FIN 1234\n',
            b'RDY 0\n',
            b'REQ 1234 0\n',
            b'FIN 1234\n',
            b'RDY 2\n',
        ]
        assert conn1.stream.write.call_args_list == [call(arg) for arg in expected_args]

        expected_args = [
            b'SUB test test\n',
            b'RDY 1\n',
            b'RDY 0\n',
            b'RDY 2\n'
        ]
        assert conn2.stream.write.call_args_list == [call(arg) for arg in expected_args]

    @gen_test
    def test_backoff_requeue_recovery(self):
        r = get_reader(max_in_flight=2)
        conn = get_conn(r)
        msg = send_message(conn)

        msg.trigger(event.FINISH, message=msg)
        assert r.backoff_block is False
        assert r.backoff_timer.get_interval() == 0

        msg = send_message(conn)

        # go into backoff
        msg.trigger(event.REQUEUE, message=msg)
        assert r.backoff_block is True
        assert r.backoff_timer.get_interval() > 0

        yield gen.sleep(r.backoff_timer.get_interval() + 0.05)

        assert r.backoff_block is False
        assert r.backoff_timer.get_interval() != 0

        msg = send_message(conn)

        # This should not move out of backoff (since backoff=False)
        msg.trigger(event.REQUEUE, message=msg, backoff=False)
        assert r.backoff_block is True
        assert r.backoff_timer.get_interval() != 0

        # elapse time
        yield gen.sleep(r.backoff_timer.get_interval() + 0.05)

        assert r.backoff_block is False
        assert r.backoff_timer.get_interval() != 0

        # this should move out of backoff state
        msg = send_message(conn)
        msg.trigger(event.FINISH, message=msg)
        assert r.backoff_block is False
        assert r.backoff_timer.get_interval() == 0

        print(conn.stream.write.call_args_list)

        expected_args = [
            b'SUB test test\n',
            b'RDY 1\n',
            b'RDY 2\n',
            b'FIN 1234\n',
            b'RDY 0\n',
            b'REQ 1234 0\n',
            b'RDY 1\n',
            b'RDY 0\n',
            b'REQ 1234 0\n',
            b'RDY 1\n',
            b'RDY 2\n',
            b'FIN 1234\n'
        ]
        assert conn.stream.write.call_args_list == [call(arg) for arg in expected_args]


@patch("tornado.ioloop.IOLoop.current")
def test_backoff_hard(mock_ioloop_current):
    mock_ioloop = get_ioloop()
    mock_ioloop_current.return_value = mock_ioloop
    r = get_reader()
    conn = get_conn(r)

    expected_args = [b'SUB test test\n', b'RDY 1\n', b'RDY 5\n']

    num_fails = 0
    fail = True
    last_timeout_time = 0
    for i in range(50):
        msg = send_message(conn)

        if fail:
            msg.trigger(event.REQUEUE, message=msg)
            num_fails += 1

            expected_args.append(b'RDY 0\n')
            expected_args.append(b'REQ 1234 0\n')
        else:
            msg.trigger(event.FINISH, message=msg)
            num_fails -= 1

            expected_args.append(b'RDY 0\n')
            expected_args.append(b'FIN 1234\n')

        assert r.backoff_block is True
        assert r.backoff_timer.get_interval() > 0
        assert mock_ioloop.add_timeout.called

        timeout_args, timeout_kwargs = mock_ioloop.add_timeout.call_args
        if timeout_args[0] != last_timeout_time:
            timeout_args[1]()
            last_timeout_time = timeout_args[0]
        assert r.backoff_block is False
        expected_args.append(b'RDY 1\n')

        fail = True
        if random.random() < 0.3 and num_fails > 1:
            fail = False

    for i in range(num_fails - 1):
        msg = send_message(conn)

        msg.trigger(event.FINISH, message=msg)
        expected_args.append(b'RDY 0\n')
        expected_args.append(b'FIN 1234\n')
        timeout_args, timeout_kwargs = mock_ioloop.add_timeout.call_args
        if timeout_args[0] != last_timeout_time:
            timeout_args[1]()
            last_timeout_time = timeout_args[0]
        expected_args.append(b'RDY 1\n')

    msg = send_message(conn)

    msg.trigger(event.FINISH, message=msg)
    expected_args.append(b'RDY 5\n')
    expected_args.append(b'FIN 1234\n')

    assert r.backoff_block is False
    assert r.backoff_timer.get_interval() == 0

    for i, f in enumerate(conn.stream.write.call_args_list):
        print("%d: %s" % (i, f))
    assert conn.stream.write.call_args_list == [call(arg) for arg in expected_args]


@patch("tornado.ioloop.IOLoop.current")
def test_backoff_many_conns(mock_ioloop_current):
    mock_ioloop = get_ioloop()
    mock_ioloop_current.return_value = mock_ioloop
    r = get_reader()

    num_conns = 5
    conns = []
    for i in range(num_conns):
        conn = get_conn(r)
        conn.expected_args = [b'SUB test test\n', b'RDY 1\n']
        conn.last_exp_rdy = b'RDY 1\n'
        conn.fails = 0
        conns.append(conn)

    fail = True
    total_fails = 0
    last_timeout_time = 0
    conn = random.choice(conns)
    for i in range(50):
        msg = send_message(conn)

        if r.backoff_timer.get_interval() == 0:
            add_exp_rdy(conn, b'RDY 1\n')

        if fail or not conn.fails:
            msg.trigger(event.REQUEUE, message=msg)
            total_fails += 1
            conn.fails += 1

            for c in conns:
                add_exp_rdy(c, b'RDY 0\n')
            conn.expected_args.append(b'REQ 1234 0\n')
        else:
            msg.trigger(event.FINISH, message=msg)
            total_fails -= 1
            conn.fails -= 1

            for c in conns:
                add_exp_rdy(c, b'RDY 0\n')
            conn.expected_args.append(b'FIN 1234\n')

        assert r.backoff_block is True
        assert r.backoff_timer.get_interval() > 0
        assert mock_ioloop.add_timeout.called

        timeout_args, timeout_kwargs = mock_ioloop.add_timeout.call_args
        if timeout_args[0] != last_timeout_time:
            conn = timeout_args[1]()
            last_timeout_time = timeout_args[0]
        assert r.backoff_block is False
        add_exp_rdy(conn, b'RDY 1\n')

        fail = True
        if random.random() < 0.3 and total_fails > 1:
            fail = False

    while total_fails:
        print("%r: %d fails (%d total_fails)" % (conn, conn.fails, total_fails))

        if not conn.fails:
            # force an idle connection
            for c in conns:
                if c.rdy > 0:
                    c.last_msg_timestamp = time.time() - 60
                    add_exp_rdy(c, b'RDY 0\n')
            conn = r._redistribute_rdy_state()
            add_exp_rdy(conn, b'RDY 1\n')
            continue

        msg = send_message(conn)

        msg.trigger(event.FINISH, message=msg)
        total_fails -= 1
        conn.fails -= 1

        if total_fails > 0:
            for c in conns:
                add_exp_rdy(c, b'RDY 0\n')
        else:
            for c in conns:
                add_exp_rdy(c, b'RDY 1\n')

        conn.expected_args.append(b'FIN 1234\n')

        timeout_args, timeout_kwargs = mock_ioloop.add_timeout.call_args
        if timeout_args[0] != last_timeout_time:
            conn = timeout_args[1]()
            last_timeout_time = timeout_args[0]

        if total_fails > 0:
            add_exp_rdy(conn, b'RDY 1\n')

    assert r.backoff_block is False
    assert r.backoff_timer.get_interval() == 0

    for c in conns:
        for i, f in enumerate(c.stream.write.call_args_list):
            print("%d: %s" % (i, f))
        assert c.stream.write.call_args_list == [call(arg) for arg in c.expected_args]


@patch("tornado.ioloop.IOLoop.current")
def test_backoff_conns_disconnect(mock_ioloop_current):
    mock_ioloop = get_ioloop()
    mock_ioloop_current.return_value = mock_ioloop
    r = get_reader()

    num_conns = 5
    conns = []
    for i in range(num_conns):
        conn = get_conn(r)
        conn.expected_args = [b'SUB test test\n', b'RDY 1\n']
        conn.last_exp_rdy = b'RDY 1\n'
        conn.fails = 0
        conns.append(conn)

    fail = True
    total_fails = 0
    last_timeout_time = 0
    conn = random.choice(conns)
    for i in range(50):
        if i % 5 == 0:
            if len(r.conns) == num_conns:
                conn.trigger(event.CLOSE, conn=conn)
                conns.remove(conn)
                if conn.rdy and r.backoff_timer.get_interval():
                    assert r.need_rdy_redistributed
                conn = r._redistribute_rdy_state()
                if not conn:
                    conn = random.choice(conns)
                else:
                    add_exp_rdy(conn, b'RDY 1\n')
                continue
            else:
                c = get_conn(r)
                c.expected_args = [b'SUB test test\n']
                c.last_exp_rdy = b'RDY 0\n'
                c.fails = 0
                conns.append(c)

        msg = send_message(conn)

        if r.backoff_timer.get_interval() == 0:
            add_exp_rdy(conn, b'RDY 1\n')

        if fail or not conn.fails:
            msg.trigger(event.REQUEUE, message=msg)
            total_fails += 1
            conn.fails += 1

            for c in conns:
                add_exp_rdy(c, b'RDY 0\n')
            conn.expected_args.append(b'REQ 1234 0\n')
        else:
            msg.trigger(event.FINISH, message=msg)
            total_fails -= 1
            conn.fails -= 1

            for c in conns:
                add_exp_rdy(c, b'RDY 0\n')
            conn.expected_args.append(b'FIN 1234\n')

        assert r.backoff_block is True
        assert r.backoff_timer.get_interval() > 0
        assert mock_ioloop.add_timeout.called

        timeout_args, timeout_kwargs = mock_ioloop.add_timeout.call_args
        if timeout_args[0] != last_timeout_time:
            conn = timeout_args[1]()
            last_timeout_time = timeout_args[0]
        assert r.backoff_block is False
        add_exp_rdy(conn, b'RDY 1\n')

        fail = True
        if random.random() < 0.3 and total_fails > 1:
            fail = False

    while total_fails:
        print("%r: %d fails (%d total_fails)" % (conn, conn.fails, total_fails))

        msg = send_message(conn)

        msg.trigger(event.FINISH, message=msg)
        total_fails -= 1
        conn.fails -= 1

        if total_fails > 0:
            for c in conns:
                add_exp_rdy(c, b'RDY 0\n')
        else:
            for c in conns:
                add_exp_rdy(c, b'RDY 1\n')

        conn.expected_args.append(b'FIN 1234\n')

        timeout_args, timeout_kwargs = mock_ioloop.add_timeout.call_args
        if timeout_args[0] != last_timeout_time:
            conn = timeout_args[1]()
            last_timeout_time = timeout_args[0]

        if total_fails > 0:
            add_exp_rdy(conn, b'RDY 1\n')

    assert r.backoff_block is False
    assert r.backoff_timer.get_interval() == 0

    for c in conns:
        for i, f in enumerate(c.stream.write.call_args_list):
            print("%d: %s" % (i, f))
        assert c.stream.write.call_args_list == [call(arg) for arg in c.expected_args]


def add_exp_rdy(conn, rdy):
    if conn.last_exp_rdy != rdy:
        conn.expected_args.append(rdy)
        conn.last_exp_rdy = rdy



from __future__ import absolute_import
import time
import unittest

import tornado.gen

import nsq

from .test_reader import IntegrationBase


class WriterUnitTest(unittest.TestCase):

    def setUp(self):
        super(WriterUnitTest, self).setUp()

    def test_constructor(self):
        name = 'test'
        reconnect_interval = 10.0
        writer = nsq.Writer(nsqd_tcp_addresses=['127.0.0.1:4150'],
                            reconnect_interval=reconnect_interval,
                            name=name)
        self.assertEqual(writer.name, name)
        self.assertEqual(0, len(writer.conn_kwargs))
        self.assertEqual(writer.reconnect_interval, reconnect_interval)

    def test_bad_writer_arguments(self):
        bad_options = dict(foo=10)

        self.assertRaises(
            AssertionError,
            nsq.Writer,
            nsqd_tcp_addresses=['127.0.0.1:4150'],
            reconnect_interval=15.0,
            name='test', **bad_options)


class WriterIntegrationTest(IntegrationBase):
    identify_options = {
        'user_agent': 'sup',
        'heartbeat_interval': 10,
        'output_buffer_size': 4096,
        'output_buffer_timeout': 50
    }

    nsqd_command = ['nsqd', '--verbose']

    def test_writer_mpub_one(self):
        topic = 'test_writer_mpub_%s' % time.time()

        w = nsq.Writer(nsqd_tcp_addresses=['127.0.0.1:4150'], **self.identify_options)

        def trypub():
            w.mpub(topic, b'{"one": 1}', callback=pubcb)

        def pubcb(conn, data):
            if isinstance(data, nsq.protocol.Error):
                if 'no open connections' in str(data):
                    self.io_loop.call_later(0.1, trypub)
                    return
            self.stop(data)

        self.io_loop.call_later(0.1, trypub)
        result = self.wait()
        print(str(result))
        assert not isinstance(result, Exception)

    def test_writer_await_pub(self):
        topic = 'test_writer_mpub_%s' % time.time()

        w = nsq.Writer(nsqd_tcp_addresses=['127.0.0.1:4150'], **self.identify_options)

        @tornado.gen.coroutine
        def trypub():
            yield w.pub(topic, b'{"one": 1}')
            yield w.pub(topic, b'{"two": 2}')
            self.stop("OK")

        self.io_loop.call_later(0.1, trypub)
        result = self.wait()
        print(str(result))
        assert result == "OK"
